/* real time linux task for parallel port stimulator */

#define PORT PAR

#include <linux/errno.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <asm/io.h> 
#include <rtl.h>
#include <rtl_sched.h>
#include <rtl_fifo.h>
#include <time.h>
#include <pthread.h>
#include "control.h"

pthread_t mytask_ton[8], mytask_toff[8];
pthread_t mytask_pon[8], mytask_poff[8];
struct my_msg_struct msg;

/* ........................................... */
/* functions to set data bits of parallel port */
/* ........................................... */
void *train_on(void *t) {
  hrtime_t now;
    while(1) {
      pthread_wait_np();
      now=gethrtime();
      pthread_wakeup_np(mytask_pon[(int)t]);
      pthread_make_periodic_np(mytask_pon[(int)t], 
			       now,
			       HRTICKS_PER_SEC/(float)1000.0*(float)msg.pulse_l[(int)t]);
      pthread_wakeup_np(mytask_poff[(int)t]);
      pthread_make_periodic_np(mytask_poff[(int)t], 
			       now+HRTICKS_PER_SEC/(float)1000.0*(float)msg.pulse_d[(int)t],
			       HRTICKS_PER_SEC/(float)1000.0*(float)msg.pulse_l[(int)t]);
    }
}

void *train_off(void *t) {
  int aus[8];
  aus[0]=0xFE; aus[1]=0xFD; aus[2]=0xFB; aus[3]=0xF7;
  aus[4]=0xEF; aus[5]=0xDF; aus[6]=0xBF; aus[7]=0x7F;
  while(1) {
    pthread_wait_np();
    pthread_suspend_np(mytask_pon[(int)t]);
    pthread_suspend_np(mytask_poff[(int)t]);
    outb(aus[(int)t] & inb(PORT), PORT);
  }
}

void *pulse_on(void *t) {
  int an[8];
  an[0]=0x01; an[1]=0x02; an[2]=0x04; an[3]=0x08;
  an[4]=0x10; an[5]=0x20; an[6]=0x40; an[7]=0x80;
  while(1) {
    pthread_wait_np();
    outb(an[(int)t] | inb(PORT), PORT);
  }
}

void *pulse_off(void *t) {
  int aus[8];
  aus[0]=0xFE; aus[1]=0xFD; aus[2]=0xFB; aus[3]=0xF7;
  aus[4]=0xEF; aus[5]=0xDF; aus[6]=0xBF; aus[7]=0x7F;
  while(1) {
    pthread_wait_np();
    outb(aus[(int)t] & inb(PORT), PORT);
  }
}

int my_handler(unsigned int fifo) {
  int       err,i;
  int       aus[8];
  hrtime_t  now;
  aus[0]=0xFE; aus[1]=0xFD; aus[2]=0xFB; aus[3]=0xF7;
  aus[4]=0xEF; aus[5]=0xDF; aus[6]=0xBF; aus[7]=0x7F;

  /* FIFO 8 transfers commands to control the driver */
  while ((err=rtf_get(10, &msg, sizeof(msg))) == sizeof(msg)) {
    switch(msg.command) {
    case START_TASK:
      for (i=0; i<8; i++) {
	if (msg.is_active[i]) {
	  now=gethrtime();
	  pthread_wakeup_np(mytask_ton[i]);
	  pthread_make_periodic_np(mytask_ton[i], 
				   now+HRTICKS_PER_SEC/(float)10.0
				   +HRTICKS_PER_SEC/(float)1000.0*(float)msg.phase[i], 
				   HRTICKS_PER_SEC/(float)1000.0*(float)msg.train_l[i]);
	  pthread_wakeup_np(mytask_toff[i]);
	  pthread_make_periodic_np(mytask_toff[i], 
				   now+HRTICKS_PER_SEC/(float)10.0
				   +HRTICKS_PER_SEC/(float)1000.0*(float)msg.phase[i]
				   +HRTICKS_PER_SEC/(float)1000.0*(float)msg.train_d[i], 
				   HRTICKS_PER_SEC/(float)1000.0*(float)msg.train_l[i]);
	}
      }
      break;
    case STOP_TASK:
      for (i=0; i<8; i++) {
	if (msg.is_active[i]) {
	  pthread_suspend_np(mytask_ton[i]);
	  pthread_suspend_np(mytask_toff[i]);
	  pthread_suspend_np(mytask_pon[i]);
	  pthread_suspend_np(mytask_poff[i]);
	  outb(aus[i] & inb(PORT), PORT);
	}
      }
      break;
    default:
      return -EINVAL;
    }
  }
  if (err!=0) {
    return -EINVAL;
  }
  return 0;
}

int init_module(void) {
  pthread_attr_t     attr;
  struct sched_param sched_param;
  int                i;

  /* PARALLEL reset */
  outb(0, PORT);

  /* handler for communication */
  rtf_create(10,1024);  /* communication with LINUX program */
  rtf_create_handler(10, &my_handler); /* handler that handels communication with LINUX program */

  /* implementation  for 8 channels */
  pthread_attr_init(&attr);
  sched_param.sched_priority = 2;
  pthread_attr_setschedparam(&attr, &sched_param);
  for (i=0; i<8; i++) {
    pthread_create(&mytask_pon[i],  &attr, pulse_on, (void *)i);
    pthread_create(&mytask_poff[i],  &attr, pulse_off, (void *)i);
    pthread_create(&mytask_ton[i],  &attr, train_on, (void *)i);
    pthread_create(&mytask_toff[i],  &attr, train_off, (void *)i);
    pthread_setfp_np(mytask_pon[i],1);
    pthread_setfp_np(mytask_poff[i],1);
    pthread_setfp_np(mytask_ton[i],1);
    pthread_setfp_np(mytask_toff[i],1);
  }
  return 0;
}

void cleanup_module(void) {
  int i;
  rtf_destroy(10);
  for (i=0; i<8; i++) {
    pthread_delete_np(mytask_ton[i]);
    pthread_delete_np(mytask_toff[i]);
    pthread_delete_np(mytask_pon[i]);
    pthread_delete_np(mytask_poff[i]);
  }
  /* PARALLEL reset */
  outb(0, PORT);
}







